import json

from python_qa.logging.logging import Logging
import logging


class BaseLogging:
    log_level = logging.NOTSET
    log_format = None

    def log(self, data):
        try:
            content = json.dumps(data.json(), indent=4, ensure_ascii=False)
        except ValueError:
            content = data.text
        if type(data.request.body) is str:
            body = json.dumps(json.loads(data.request.body), indent=4, ensure_ascii=False)
        else:
            body = "-"
        Logging.logger.log(
            self.log_level, self.log_format.format(data=data, body=body, content=content)
        )


class LoggingResponseInfo(BaseLogging):
    log_format = (
        "Request: {data.request.method} {data.url} "
        "headers {data.request.headers}\n"
        "{body}\n"
        "Response: {data.status_code}\n"
        "{content}"
    )
    log_level = logging.INFO

    def run(self, response, *args, **kwargs):
        self.log(response)



import logging


class Logging:
    LEVEL = logging.INFO
    logger: logging.Logger = None

    @staticmethod
    def set_log_level(level):
        Logging.LEVEL = level
        Logging.logger = logging.getLogger("python-qa")
        Logging.logger.setLevel(level)


Logging.set_log_level(Logging.LEVEL)



import inspect
import typing
import random as rnd
import uuid
from dataclasses import is_dataclass, MISSING
import datetime

import attr
from faker import Faker

from ..utils.data import DataGenerate
from ..logging.logging import Logging
from ..utils.classes import is_attrs_class, is_union_type, has_args, is_tuple, is_sequence
from ..types.base_types import *

logger = Logging.logger

T = typing.TypeVar("T")


class Randomer:
    _types = {}

    def get_func(self, t: T):
        func = self._types.get(t)
        if func:
            return func
        raise TypeError(f"Not found func for random {t} type data")

    def run_func(self, t: T, **kwargs):
        func = self.get_func(t)
        if inspect.getfullargspec(func).varkw:
            logger.debug(f"Parameters {kwargs} are passed to func")
            res = func(**kwargs)
        else:
            res = func()
        logger.debug(f"Data random for type {t}: {res}")
        return res

    def add_type(self, t: T, func: typing.Callable[[], T]):
        if t in self._types:
            logger.warning(f"Existing type {t} will be overriding")
        logger.debug(f"New type {t} with function {func} registering")
        self._types[t] = func

    def add_types(self, types: typing.Dict[T, typing.Callable[[], T]]):
        logger.debug(f"New types registered: {types}")
        self._types.update(types)

    def add_predefined(self, **kwargs):
        fake = Faker(**kwargs)
        types = {
            str: lambda: fake.sentence(2)[:-1],
            int: fake.pyint,
            float: fake.pyfloat,
            bool: fake.pybool,
            list: fake.pylist,
            Date: fake.date,
            Address: fake.address,
            Country: fake.country,
            CountryCode: fake.country_code,
            PostCode: fake.postcode,
            StreetAddress: fake.street_address,
            CarNumber: fake.license_plate,
            BBAN: fake.bban,
            IBAN: fake.iban,
            SWIFT11: fake.swift11,
            SWIFT8: fake.swift8,
            CreditCardExpire: fake.credit_card_expire,
            CreditCardNumber: fake.credit_card_number,
            CreditCardProvider: fake.credit_card_provider,
            CreditCardSecurityCode: fake.credit_card_security_code,
            CompanyName: fake.company,
            CompanySuffix: fake.company_suffix,
            FileName: fake.file_name,
            Ipv4: fake.ipv4,
            Ipv6: fake.ipv6,
            MacAddress: fake.mac_address,
            UserAgent: fake.user_agent,
            URI: fake.uri,
            Email: fake.email,
            Username: fake.user_name,
            FirstName: fake.first_name_male,
            LastName: fake.last_name_male,
            MiddleName: fake.middle_name_male,
            Password: fake.password,
            Patronymic: fake.middle_name_male,
            PhoneNumber: fake.phone_number,
            Job: fake.job,
            Paragraph: fake.paragraph,
            BusinessInn: fake.businesses_inn,
            IndividualInn: fake.individuals_inn,
            BusinessOgrn: fake.businesses_ogrn,
            IndividualOgrn: fake.individuals_ogrn,
            Kpp: fake.kpp,
            TimeStamp: lambda: fake.unix_time(
                start_datetime=(
                    datetime.date.today() - datetime.timedelta(days=30)
                )
            ),
            MongoId: DataGenerate.mongo_id,
            Uuid: lambda: str(uuid.uuid4()),
        }
        self.add_types(types)

    def random_object(self, t: T, fields: typing.List[str] = [],
                      ignore_fields: bool = False,
                      save_defaults: bool = False,
                      **kwargs
                      ) -> T:
        data = {}
        if t in self._types:
            return self.run_func(t, **kwargs)
        elif is_attrs_class(t):
            for field in attr.fields(t):
                f_name = field.name
                has_default = field.default is not attr.NOTHING if save_defaults else None
                if f_name in kwargs:
                    data[f_name] = kwargs[f_name]
                    continue
                if (
                        (f_name in fields and ignore_fields) or
                        (save_defaults and has_default) or
                        (f_name not in fields and not ignore_fields and fields)
                ):
                    data[f_name] = field.default if has_default else None
                    continue
                data[f_name] = self.random_object(
                    field.type,
                    fields,
                    ignore_fields,
                    save_defaults,
                    **kwargs
                )
            logger.debug(f"Random data is generate for attrs type {data}")
            return t(**data)
        elif is_dataclass(t):
            for field in fields(t):
                f_name = field.name
                has_default = (
                    field.default is not MISSING
                    or field.default_factory is not MISSING
                ) if save_defaults else None
                if f_name in kwargs:
                    data[f_name] = kwargs[f_name]
                    continue
                if (
                        (f_name in fields and ignore_fields) or
                        (save_defaults and has_default) or
                        (f_name not in fields and not ignore_fields and fields)
                ):
                    data[f_name] = None
                    if has_default:
                        data[f_name] = (
                            field.default
                            if field.default != MISSING
                            else field.default_factory()
                        )
                    continue
                data[f_name] = self.random_object(
                    field.type,
                    fields,
                    ignore_fields,
                    save_defaults,
                    **kwargs
                )
            logger.debug(f"Random data is generate for dataclass type {data}")
            return t(**data)
        elif is_union_type(t) and has_args(t):
            return self.random_object(
                rnd.choice(t.__args__),
                fields,
                ignore_fields,
                save_defaults,
                **kwargs
            )
        elif is_tuple(t) and has_args(t):
            return (
                self.random_object(
                    t.__args__[0],
                    fields,
                    ignore_fields,
                    save_defaults,
                    **kwargs
                ),
            )
        elif is_sequence(t) and has_args(t):
            return [
                self.random_object(
                    rnd.choice(t.__args__),
                    fields,
                    ignore_fields,
                    save_defaults,
                    **kwargs
                )
            ]
        return None

    object = random_object



import functools
import json

import attr
import cattr

from python_qa.utils.func import func_parameters
from python_qa.utils.classes import is_attrs_class
from requests import Response

from python_qa.utils.random import logger

from python_qa.utils.func import func_args_dict


def throw_response_missmatch(
        fields: dict, resp: Response, exception: Exception
):
    raise TypeError(
        f"Parameters miss-matched. "
        f"First level: "
        f"\n\tExpect: {sorted(fields)}"
        f"\n\tActual: {sorted(resp.json().keys())}"
        f"\n\tOriginal exception: {exception}"
    )


class Decorators:
    @staticmethod
    def request_typing(response_type=None, client_error_type=None):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                f_args = func_args_dict(func, *args, **kwargs)
                data = f_args.get("data", None)
                params = f_args.get("params", None)
                if is_attrs_class(data):
                    data = cattr.unstructure(data)
                    f_args["data"] = json.dumps(data)
                elif isinstance(data, list) and is_attrs_class(data[0]):
                    list_data = []
                    for d in data:
                        list_data.append(cattr.unstructure(d))
                    f_args["data"] = json.dumps(list_data)
                if is_attrs_class(params):
                    f_args["params"] = cattr.unstructure(params)
                args = f_args.values()
                resp = func(*args, **kwargs)
                if response_type and resp.status_code // 100 == 2:
                    untyped = resp.json()
                    try:
                        if isinstance(untyped, list):
                            typed = []
                            for item in untyped:
                                typed.append(cattr.structure(item, response_type))
                            resp.typed = typed
                        else:
                            resp.typed = cattr.structure(untyped, response_type)
                    except (TypeError, ValueError, KeyError) as exp:
                        fields = attr.fields_dict(response_type)
                        throw_response_missmatch(fields, resp, exp)
                elif client_error_type and resp.status_code // 100 == 4:
                    resp.typed = cattr.structure(resp.json(), client_error_type)
                return resp
            return wrapper
        return decorator

    @staticmethod
    def step(title: str = None):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                params = func_parameters(func, *args, **kwargs)
                try:
                    params_title = title.format(*args, **params)
                except KeyError:
                    params_title = title
                logger.info("Executing: " + params_title)
                return func(*args, **kwargs)

            return wrapper

        return decorator


deco = Decorators



import os

import git
import pytest


class PyTestReRunPlugin:
    diff = []

    def __init__(self, main_branch, diff_branch):
        self.main_branch = main_branch
        self.diff_branch = diff_branch
        self.repo = git.Repo(os.getcwd())
        self.repo.git.fetch()
        self.collect_diff()

    def collect_diff(self):
        diff_branch = (
            self.diff_branch
            if self.diff_branch != "current" else self.repo.active_branch.name
        )
        diff = self.repo.git.diff(
            self.main_branch + '..' + diff_branch, name_only=True
        ).split("\n")
        self.diff = [os.path.abspath(i) for i in diff]
        print(diff)

    @pytest.hookimpl(tryfirst=True)
    def pytest_ignore_collect(self, path, config):
        if os.path.isdir(path):
            return False
        elif not self.diff:
            return False
        return path not in self.diff



from dataclasses import dataclass
from enum import Enum
from random import choice
from typing import List, TypeVar

import attr
from python_qa.logging.logging import Logging

T = TypeVar("T")

def str_type(name: str):
    return type(name, (str,), {})


def int_type(name: str):
    return type(name, (int,), {})


Address = str_type("Address")
BBAN = str_type("BBAN")
CarNumber = str_type("CarNumber")
Country = str_type("Country")
CountryCode = str_type("CountryCode")
CreditCardExpire = str_type("CreditCardExpire")
CreditCardNumber = str_type("CreditCardNumber")
CreditCardProvider = str_type("CreditCardProvider")
CreditCardSecurityCode = str_type("CreditCardSecurityCode")
CompanyName = str_type("CompanyName")
CompanySuffix = str_type("CompanySuffix")
Date = str_type("Date")
Email = str_type("Email")
FileName = str_type("FileName")
FirstName = str_type("FirstName")
PostCode = str_type("PostCode")
StreetAddress = str_type("StreetAddress")
IBAN = str_type("IBAN")
SWIFT11 = str_type("SWIFT11")
SWIFT8 = str_type("SWIFT8")
Ipv4 = str_type("Ipv4")
Ipv6 = str_type("Ipv6")
MacAddress = str_type("MacAddress")
UserAgent = str_type("UserAgent")
URI = str_type("URI")
Username = str_type("UserName")
LastName = str_type("LastName")
MiddleName = str_type("MiddleName")
Paragraph = str_type("Paragraph")
Password = str_type("Password")
Patronymic = str_type("Patronymic")
PhoneNumber = str_type("PhoneNumber")
Job = str_type("Job")
BusinessInn = str_type("BusinessInn")
IndividualInn = str_type("IndividualInn")
BusinessOgrn = str_type("BusinessOgrn")
IndividualOgrn = str_type("BusinessOgrn")
Kpp = str_type("Kpp")
TimeStamp = int_type("TimeStamp")
MongoId = str_type("MongoId")
Uuid = str_type("Uuid")


class BaseEnum(Enum):
    @classmethod
    def all(cls):
        return [i for i in cls]

    @classmethod
    def random(cls):
        return choice(cls.all())


@dataclass
class NamedData:
    code: str
    text: str
    digital_code: int = None


class NamedEnum(Enum):
    @classmethod
    def all(cls) -> List[NamedData]:
        return [i for i in cls]

    @classmethod
    def all_code(cls):
        return [i.value.code for i in cls]

    @classmethod
    def all_text(cls):
        return [i.value.text for i in cls]

    @classmethod
    def random(cls):
        return choice(cls.all())

    @classmethod
    def random_code(cls):
        return choice(cls.all_code())

    @classmethod
    def random_text(cls):
        return choice(cls.all_text())

    @classmethod
    def dict(cls) -> dict:
        d = {}
        for i in cls:
            d[i.value.code] = i.value.text
        return d

    @classmethod
    def text_by_code(cls, code: str) -> str:
        return cls.dict().get(code)

    @classmethod
    def code_by_text(cls, text: str) -> str:
        for k, v in cls.dict().items():
            if v == text:
                return k
        return None

    @property
    def text(self):
        return self.value.text

    @property
    def code(self):
        return self.value.code

    @property
    def digital_code(self):
        return self.value.digital_code

class BaseType:
    @staticmethod
    def _safe_to_type(value, to_type):
        try:
            if getattr(value, "to_type", None):
                return value.to_type(to_type)
        except Exception as e:
            Logging.logger.warning(e)

    @staticmethod
    def _safe_list_to_type(value, to_type):
        try:
            type_in_list = to_type.__args__[0]
            data = [BaseType._safe_to_type(val, type_in_list) for val in value]
            if data.count(None) == len(data):
                return
            return data
        except Exception as e:
            Logging.logger.warning(e)

    def to_type(self, t: T) -> T:
        data = {}
        for field in attr.fields(t):
            value = getattr(self, field.name, None)
            if isinstance(value, list):
                data[field.name] = self._safe_list_to_type(value, field.type)
            else:
                data[field.name] = self._safe_to_type(value, field.type)
            data[field.name] = data.get(field.name) or getattr(self, field.name, None)
        return t(**data)

    def merge(self, obj):
        for field in attr.fields(type(obj)):
            value = getattr(obj, field.name, None)
            if value and field in attr.fields(type(self)):
                setattr(self, field.name, value)
        return self



from python_qa.pytest.rerun import PyTestReRunPlugin


def pytest_addoption(parser):
    parser.addoption(
        "--only-changed", action="store_true", default=False, help=""
    )
    parser.addoption(
        "--rerun-main-branch", action="store", default=None,
        help="Branch which will be used as main in diff\n"
             "(Default: origin/master)"
    )
    parser.addoption(
        "--rerun-diff-branch", action="store", default=None,
        help="Branch which will be used as diff for the main\n"
             "(Default: <current>)"
    )
    parser.addini(
        "rerun-main-branch",
        "Branch which will be used as main in diff",
        default=None,
    )
    parser.addini(
        "rerun-diff-branch",
        "Branch which will be used as diff for the main",
        default=None,
    )


def pytest_configure(config):
    if config.getoption("--only-changed"):
        main_branch = (
                config.getoption("--rerun-main-branch")
                or config.getini("rerun-main-branch")
                or "origin/master"
        )
        diff_branch = (
                config.getoption("--rerun-diff-branch")
                or config.getini("rerun-diff-branch")
                or "current"
        )
        config.pluginmanager.register(
            PyTestReRunPlugin(
                main_branch=main_branch,
                diff_branch=diff_branch
            ),
            name="pytest-rerun-instance"
        )



import collections
import inspect
import types


def represent(item):
    if isinstance(item, str):
        return f"'{item}'"
    elif isinstance(item, (bytes, bytearray)):
        return repr(type(item))
    elif isinstance(item, types.FunctionType):
        return inspect.getsource(item).strip()
    else:
        return repr(item)


def func_parameters(func, *args, **kwargs):
    parameters = {}
    args_spec = inspect.getfullargspec(func)
    args_order = list(args_spec.args)
    args_dict = dict(zip(args_spec.args, args))

    if args_spec.defaults:
        defaults = dict(
            zip(args_spec.args[len(args):], args_spec.defaults)
        )
        parameters.update(defaults)

    if args_spec.varargs:
        args_order.append(args_spec.varargs)
        varargs = args[len(args_spec.args):]
        parameters.update({args_spec.varargs: varargs} if varargs else {})

    if args_spec.args and args_spec.args[0] in ["cls", "self"]:
        args_dict.pop(args_spec.args[0], None)

    if kwargs:
        args_order.extend(list(kwargs.keys()))
        parameters.update(kwargs)

    parameters.update(args_dict)
    items = parameters.items()
    sorted_items = sorted(
        map(lambda k: (k[0], represent(k[1])), items),
        key=lambda x: args_order.index(x[0]),
    )

    return collections.OrderedDict(sorted_items)


def func_args_dict(func, *args):
    args_spec = inspect.getfullargspec(func)
    return dict(zip(args_spec.args, args))



from typing import Union, Tuple, List, Sequence, MutableSequence


def is_attrs_class(cls):
    return getattr(cls, "__attrs_attrs__", None) is not None


def has_args(cls):
    return getattr(cls, "__args__", None) is not None


def is_tuple(t):
    return (
            t in (Tuple, tuple) or
            (getattr(t, "__origin__", None) is tuple)
    )


def is_sequence(t):
    return (
            t in (List, list, Sequence, MutableSequence)
            or (getattr(t, "__origin__", None) is list)
    )


def is_union_type(obj):
    return (
        obj is Union or getattr(obj, "__origin__", None) is Union
    )


class StructDict:
    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)
        for key, value in self.__dict__.items():
            if isinstance(value, dict):
                self.__dict__[key] = StructDict(**value)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="python-qa",
    version="0.0.5",
    author="Starkov E.G.",
    author_email="Starkov.Ev.Ge@gmail.com",
    license="Apache 2.0",
    description="Simplify the QA of your product",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Starkov-EG/python-qa",
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
    entry_points={'pytest11': ['python_qa = python_qa.conftest']},
)



import binascii
import os
import time


class DataGenerate:
    @staticmethod
    def mongo_id() -> str:
        timestamp = int(time.time())
        other_part = binascii.b2a_hex(os.urandom(8)).decode("ascii")
        return f"{timestamp:x}{other_part}"



import inspect
from typing import Union, List

from requests import Response

from python_qa.logging.logging import Logging


def verify_response(resp: Response, ok_status: Union[int, List[int]] = 200) -> Response:
    func = inspect.stack()[1][3]
    if isinstance(ok_status, int):
        ok_status = [ok_status]
    if resp.status_code not in ok_status:
        raise ValueError(
            f"Verified response: function {func} failed:"
            f"\nResponse code: {resp.status_code}, expected: {ok_status}"
            f"\nResponse data: {resp.content}"
        )
    else:
        Logging.logger.info(
            f"Verified response: function {func} code {resp.status_code}"
        )
    return resp


vr = verify_response
